#!/usr/bin/perl

use strict;
use warnings;

use Data::Dumper ();
use Getopt::Long ();

my $SHIM_SIZE = 8192;
my $SHIM_BEFORE = 2048;
my $SHIM_AFTER = $SHIM_SIZE - $SHIM_BEFORE;
my $BOOT_SIZE = 4096;

my $Dry_Run = 0;
my $Debug = 0;
my $Table_Dir = "/etc/xen/tables";
my $Lvm_Vg = "dom0.root";
my $Target_Size = 40 * 1024 * 1024 * 2; # 40GiB
my $Boot_Partition_Number = 99;
my $Xencfg;

my @Tables_Created;
my @Shims_Created;
my @Device_Nodes_Created;

#
# Basically just system, but allows for dry runs and reporting the
# exact command executed on STDERR.
#
sub run {

    if ($Dry_Run || ($Debug >= 1)) {
        local $Data::Dumper::Terse = 1;
        print STDERR "+system: ", Data::Dumper->Dump(\@_);
        $? = 0;
    }
    if (!$Dry_Run) {
        system(@_);
    }

    return $?;
}

#
# Returns an ARRAY reference of the lines in the Xen configuration.
#
sub get_cfglines {

    if (!open XENCFG, "< $Xencfg") {
        die("Could not open $Xencfg");
    }

    my (@lines, $line);
    while (($line = <XENCFG>)) {
        push @lines, $line;
    }

    close XENCFG;
    return \@lines;
}

#
# Given the lines of the configuration file (passed as an ARRAY
# reference), returns the configured hostname.
#
sub extract_hostname {

    my ($lines,) = @_;

    foreach my $line (@$lines) {
        if ($line =~ /^\s*name\s*=\s*('|")(.*)\g1/) {
            return $2;
        }
    }

    die("Could not extract hostname from the Xen configuration");
}

#
# Given the lines of the configuration file (passed as an ARRAY
# reference), returns a list indicating:
#
# [0] index of the first cfgfile line containing the disk list
# [1] index of the last cfgfile line containing the disk list
# [2] the text of the disk list, as a SCALAR Reference
#
sub extract_disklist {

    my ($lines,) = @_;

    my ($first, $last, @disklist);
    for (my $ii = 0; $ii < @$lines; $ii++) {
        my $line = $lines->[$ii];
        if (!defined($first)) {
            if ($line =~ /^\s*disk\s*=\s*\[(.*)$/) {
                $line = $1;
                $first = $ii;
            } else {
                next;
            }
        }
        if (!defined($last) && ($line =~ /^([^#]*)\]/)) {
            $line = $1;
            $last = $ii;
        }

        $line =~ s/\s*#.*$//;
        if ($line !~ /^\s*$/) {
            push @disklist, $line;
        }
        if (defined($last)) {
            last;
        }
    }

    if ($Debug >= 2) {
        print STDERR Data::Dumper->Dump([$first, $last], [qw(first last)]);
    }
    if (!defined($first) || !defined($last)) {
        die("Could not extract disk list from the Xen configuration");
    }

    my $disklist = join("", @disklist);
    return ($first, $last, \$disklist);
}

#
# Identifies each quoted string within a disk list, and returns a
# list of HASH references, where each HASH reference is structured
# data about the virtual disk.
#
sub parse_disklist {

    my ($disklist,) = @_;

    # identify each quoted string within the disk list
    my @diskstrs;
    my ($start, $quote, $c);
    for (my $i = 0; $i < length($$disklist); $i++) {
        $c = substr($$disklist, $i, 1);
        if ($c eq "'" || $c eq '"') {
            if (!$quote) {
                $start = $i;
                $quote = $c;
            } elsif ($c eq $quote) {
                push @diskstrs, substr($$disklist, $start + 1, $i - $start - 1);
                undef $quote;
            }
        }
    }
    if (defined $quote) {
        die "Error parsing disk list";
    }
    if ($Debug >= 2) {
        print STDERR Data::Dumper->Dump([\@diskstrs], [qw(*diskstrs)]);
    }

    return sort {
            !defined($a->{partition}) ? -1 :
                (!defined($b->{partition}) ?  1 :
                     ($a->{partition} cmp $b->{partition}));
        } map {
            my $diskstr = $_;
            if ($Debug >= 2) {
                print STDERR Data::Dumper->Dump([$diskstr], [qw(diskstr)]);
            }
            if ($diskstr !~ /(,\s*|\bvdev=)((h|s|xv)d[a-z])(\d+)?(,|\s|$)/) {
                die "Could not identify vdev from disk string";
            }
            my $dev = $2;
            my $partition = $4;
            if ($diskstr !~ /(^([^,]*:)?|\btarget=)([^,\s]+)(,|\s)/) {
                die "Could not identify target from disk string";
            }
            my $target = $3;
            if (substr($target, 0, 1) ne '/') {
                $target = "/dev/${target}";
            }

            {dev => $dev, partition => $partition, target => $target};
        } @diskstrs;
}

#
# Re-arranges a disk list so that entries using the same non-partition
# device are grouped together.  Returns a HASH reference, where the keys
# are base devices and the values are hash references from the original
# disk list, expanded with information on the device's size.
#
sub group_disks {

    my ($disks,) = @_;

    my %groups;
    foreach my $disk (@$disks) {
        my $dev = $disk->{dev};
        my $devlist = $groups{$dev} //= [];
        if (!defined($disk->{partition})) {
            $disk->{copy} = 1;
        } else {
            my $target = $disk->{target};
            if (!$Dry_Run && ! -e $target) {
                die "Target device $target does not exist";
            }
            my $qtarget = quotemeta($target);
            my $size = `blockdev --getsz $qtarget`;
            if ($? != 0) {
                die "Failed determining size of $target";
            }
            chomp $size;
            $disk->{size} = $size;
        }
        push @$devlist, $disk;
    }
    return \%groups;
}

#
# Executes common pre-flight checks.
#
sub preflight_common {

    run("which blockdev >/dev/null 2>&1")
        and die("The 'blockdev' command is missing from your PATH");
    run("which lvcreate >/dev/null 2>&1")
        and die("The 'lvcreate' command is missing from your PATH");
    run("which dmsetup >/dev/null 2>&1")
        and die("The 'dmsetup' command is missing from your PATH");
    run("which sgdisk >/dev/null 2>&1")
        and die("The 'sgdisk' command is missing from your PATH");

    run("vgdisplay " . quotemeta($Lvm_Vg) . " >/dev/null 2>&1")
        and die("Failed to run 'vgdisplay' on the $Lvm_Vg volume group");

    (!$Dry_Run && ! -w "$Table_Dir")
        and die("Missing or un-writable $Table_Dir directory");
    (!$Dry_Run && -e "$Xencfg.bak")
        and die("Xen configuration backup already exists; refusing to clobber it");
}

#
# Executes pre-flight checks specific to a group of partitions.
#
sub preflight_group {

    my ($name) = @_;

    (-e "$Table_Dir/$name.table")
        and die("Device mapper table for $name already exists, aborting");
    (-e "/dev/$Lvm_Vg/$name-shim")
        and die("Shim volume for $name already exists, aborting");
    (-e "/dev/mapper/$name")
        and die("Device mapper node for $name already exists, aborting");
}

#
# Given a group of partitions that are part of the same device, writes
# a device mapper table gluing them together.
#
sub write_device_mapper_table {

    my ($name, $group) = @_;

    my $table = "$Table_Dir/$name.table";
    if ($Dry_Run || ($Debug >= 2)) {
        print STDERR "+table: <<-[\n";
    }
    if (!$Dry_Run) {
        if (!open TABLE, "> $table") {
            die("Could not create device mapper table for $name");
        }
        push @Tables_Created, $table;
    }

    my $p = sub {
        if ($Dry_Run || ($Debug >= 2)) {
            print STDERR " " x 8, @_;
        }
        if (!$Dry_Run) {
            print TABLE @_;
        }
    };

    my $shimpath = "/dev/$Lvm_Vg/$name-shim";
    $p->("0 $SHIM_BEFORE linear $shimpath 0\n");
    my $offset = $SHIM_BEFORE;
    foreach my $part (@$group) {
        $part->{offset} = $offset;
        $p->("$offset $part->{size} linear $part->{target} 0\n");
        $offset += $part->{size};
    }

    if ($Target_Size) {
        if ($offset > $Target_Size) {
            close TABLE if (!$Dry_Run);
            die("Existing partitions do not fit in the target size");
        }

        my $reserved = $Target_Size - $offset;
        $p->("$offset $reserved error\n");
        $offset += $reserved;
    }

    $p->("$offset $SHIM_AFTER linear $shimpath $SHIM_BEFORE\n");

    if ($Dry_Run || ($Debug >= 2)) {
        print STDERR "]\n";
    }
    if (!$Dry_Run) {
        close TABLE;
    }

    print ".. Wrote device mapper table to $table\n";
}

#
# Creates a shim LVM logical volume that will store a partition table
# and bootloader.
#
sub create_shim {

    my ($name,) = @_;

    run("lvcreate -L ${SHIM_SIZE}S" .
                " -n " . quotemeta("$name-shim") . " " . quotemeta($Lvm_Vg));
    if ($? != 0) {
        die("Could not create shim volume for $name, code $?");
    }
    push @Shims_Created, "$name-shim";
}

#
# Creates the named device mapper node, assuming the table for the
# device already exists in the correct location.
#
sub create_device {

    my ($name,) = @_;

    run("dmsetup create " . quotemeta($name) .
                " < " . quotemeta("$Table_Dir/$name.table"));
    if ($? != 0) {
        die("Could not create device mapper node for $name, code $?");
    }
    push @Device_Nodes_Created, $name;

    print ".. Created $name device mapper node\n";
}

#
# Given a group of partitions that are part of the same device, writes
# an appropriate partition table describing the partitions.
#
sub write_partition_table {

    my ($name, $group) = @_;

    run("sgdisk -Z " . quotemeta("/dev/mapper/$name"));
    if ($? != 0) {
        die("Could not initialize GPT for $name");
    }

    my (@opts, $end);
    foreach my $p (@$group) {
        my $number = $p->{partition};
        $end = $p->{offset} + $p->{size} - 1;
        push @opts, "-n $number:$p->{offset}:$end"; # create partition

        my $target = $p->{target};
        my $type = ($target =~ /swap/i) ? "8200" :
                        (($p->{target} =~ /lvm/i) ? "8e00" : "8300");
        push @opts, "-t $number:$type"; # set partition type
    }

    if ($group->[0]{dev} =~ /a$/) {
        my $offset = defined($Target_Size) ? $Target_Size : ($end + 1);
        $end = $offset + $BOOT_SIZE - 1;
        # create boot loader area marked as a BIOS boot partition
        push @opts, "-n $Boot_Partition_Number:$offset:$end",
                    "-t $Boot_Partition_Number:ef02";
    }

    run(join(" ", "sgdisk", @opts, quotemeta("/dev/mapper/$name")));
    if ($? != 0) {
        die("Could not write GPT for $name");
    }
}

#
# Re-writes the Xen configuration, making it pass the device mapper
# node as a whole device to the VM instead of passing individual partitions.
#
sub update_xencfg {

    my ($lines, $first, $last, $hostname, $groups) = @_;

    if (!$Dry_Run) {
        run("cp -p " . quotemeta($Xencfg) . " " . quotemeta("$Xencfg.bak"));
        if ($? != 0) {
            die("Could not create a backup of the old Xen configuration");
        }

        if (!open XENCFG, "> $Xencfg.new") {
            die("Could not create temp Xen configuration");
        }

        print XENCFG @{$lines}[0 .. ($first - 1)];
    }
    if ($Dry_Run || ($Debug >= 2)) {
        print STDERR "+xencfg: <<-[\n";
        print STDERR "        ...\n";
    }

    my $p = sub {
        if ($Dry_Run || ($Debug >= 2)) {
            print STDERR " " x 8, @_;
        }
        if (!$Dry_Run) {
            print XENCFG @_;
        }
    };

    $p->("# Old disk list, before processing by hvmify.pl\n");
    foreach my $line (@{$lines}[$first .. $last]) {
        $p->("#$line");
    };
    $p->("\n");
    $p->("# New disk list, created by hvmify.pl\n");
    $p->("disk = [\n");
    foreach my $dev (sort(keys(%$groups))) {
        my $entry = $groups->{$dev}[0];
        if ((@{$groups->{$dev}} == 1) && $entry->{copy}) {
            $p->("        '$entry->{target},raw,$entry->{dev},w',\n");
        } else {
            my $target = "/dev/mapper/$hostname-$entry->{dev}";
            $p->("        '$target,raw,$entry->{dev},w',\n");
        }
    }
    $p->("]\n");

    if ($Dry_Run || ($Debug >= 2)) {
        print STDERR "        ...\n";
        print STDERR "]\n";
    }
    if (!$Dry_Run) {
        print XENCFG @{$lines}[($last + 1) .. $#$lines];
        close XENCFG;

        rename("$Xencfg.new", "$Xencfg")
            or die("Could not replace old Xen configuration with the new configuration");
    }
}

#
# Cleans up objects that were created, in case there was an error.
#
sub cleanup {

    print "Cleaning up...\n";

    foreach my $table (@Tables_Created) {
        if ($Dry_Run || ($Debug >= 1)) {
            print STDERR "+unlink: " . $table . "\n";
        }
        if (!$Dry_Run) {
            unlink($table)
                or warn("Could not clean up $table");
        }
    }

    foreach my $dev_node (@Device_Nodes_Created) {
        run("dmsetup remove " . quotemeta($dev_node))
            and warn("Could not clean up $dev_node");
    }

    foreach my $shim (@Shims_Created) {
        run("lvremove -f " . quotemeta("$Lvm_Vg/$shim"))
            and warn("Could not clean up $shim");
    }

    if (-e "$Xencfg.bak") {
        unlink("$Xencfg.bak")
            or warn("Could not clean up backup Xen configuration");
    }
    if (-e "$Xencfg.new") {
        unlink("Xencfg.new")
            or warn("Could not clean up temp Xen configuration");
    }
}

sub main {

    preflight_common();

    my $lines = get_cfglines();
    my $hostname = extract_hostname($lines);
    my ($first, $last, $disklist) = extract_disklist($lines);
    my @disks = parse_disklist($disklist);
    my $groups = group_disks(\@disks);

    if ($Debug >= 2) {
        print STDERR Data::Dumper->Dump([$groups], [qw(groups)]);
    }

    eval {
        foreach my $dev (sort(keys(%$groups))) {
            my $group = $groups->{$dev};
            if ((@$group == 1) && $group->[0]{copy}) {
                print "Skipping $dev, which is passed as a whole device.\n";
                next;
            }

            print "Building device for $dev...\n";
            my $name = "$hostname-$dev";
            preflight_group($name);
            write_device_mapper_table($name, $group);
            create_shim($name);
            create_device($name);
            write_partition_table($name, $group);
        }

        print "Writing updated Xen configuration...\n";
        update_xencfg($lines, $first, $last, $hostname, $groups);
    };
    if ($@) {
        print STDERR $@;
        cleanup();
        print "Done\n";
        exit(1);
    }

    print "Done\n";
}

my $help;
my $ok = Getopt::Long::GetOptions(
    "dry-run|n" => \$Dry_Run,
    "debug|d+" => \$Debug,
    "table-dir=s" => \$Table_Dir,
    "lvm-volume-group=s" => \$Lvm_Vg,
    "target-size|s=i" => \$Target_Size,
    "boot-partition-number|B=i" => \$Boot_Partition_Number,
    "help|h" => \$help,
);
$ok &&= (@ARGV == 1);

if (!$ok) {
    print "Invalid command line arguments\n";
}
if ($help || !$ok) {
    my $basename = $0;
    $basename =~ s%^.*/%%;

    print "Usage: $basename [--dry-run] [--debug]\n";
    print "                 [--boot-partition-number=<N>] [--lvm-volume-group=<vg>]\n";
    print "                 [--table-dir=<path>] [--target-size=<sectors>] <xencfg>\n";
    exit($ok ? 0 : 1);
}

$Xencfg = $ARGV[0];
main();
