#!/usr/bin/perl

use strict;
use warnings;

use Carp qw(confess);
use Data::Dumper ();

use constant DEBUG => 0;


sub apply_ip4(&@) {
    my $op = shift;
    my @args = @_;
    if (@args == 0) {
        confess "Not enough arguments";
    }
    foreach $_ (@args) {
        if (ref($_) ne 'ARRAY') {
            confess "IP addresses should be passed as ARRAYREFs";
        } elsif (@$_ != 4) {
            die "Bad octet count '" . scalar(@$_) . "' != '4'";
        }
    }

    my $result;
    for my $i (0 .. $#{$args[0]}) {
        $result = $op->($result, map { $_->[$i] } @args);
    }
    return $result;
}


sub debug {
    my @labels;
    $#labels = $#_ - 1;
    $labels[-1] = 'ret';

    local $Data::Dumper::Indent = 0;
    my $dump = Data::Dumper->Dump([@_[1 .. $#_]], \@labels);
    $dump =~ s/;/; /g;
    print STDERR "$_[0]: $dump\n";
}


sub or_ip4 {
    my $ret = apply_ip4 { $_[0] //= []; push @{$_[0]}, ($_[1] | $_[2]); $_[0] } @_;
    debug('or_ip4', @_, $ret) if (DEBUG);
    return $ret;
}

sub and_ip4 {
    my $ret = apply_ip4 { $_[0] //= []; push @{$_[0]}, ($_[1] & $_[2]); $_[0] } @_;
    debug('and_ip4', @_, $ret) if (DEBUG);
    return $ret;
}

sub equal_ip4 {
    my $ret = apply_ip4 { $_[0] //= 1; $_[0] && ($_[1] == $_[2]) } @_;
    debug('equal_ip4', @_, $ret) if (DEBUG);
    return $ret;
}


sub main {
    #
    # Parse and validate input
    #

    if (@ARGV != 4) {
        my $basename = $0;
        $basename =~ s(^.*/)()g;
        print STDERR "Usage: $basename <network> <netmask> <address> <netmask>\n";
        exit 1;
    }

    my @network = split(/\./, $ARGV[0]);
    map { s/^$/0/ } @network;
    if (@network < 4) {
        push @network, (0,) x (4 - @network);
    }
    my @netmask1 = split(/\./, $ARGV[1]);
    my @address = split(/\./, $ARGV[2]);
    my @netmask2 = split(/\./, $ARGV[3]);

    foreach my $octet (@network, @netmask1, @address, @netmask2) {
        if ($octet !~ /^[0-9]+$/ || $octet < 0 || $octet > 255) {
            die "Bad octet '$octet'"
        }
    }

    #
    # Reconcile differing netmasks
    #
    my $netmask = or_ip4(\@netmask1, \@netmask2);

    #
    # Validate the network against the combined netmask
    #
    if (!equal_ip4(and_ip4(\@network, $netmask), \@network)) {
        die "Network has one or more host bits set";
    }

    #
    # Check whether the address is in the network
    #
    if (equal_ip4(and_ip4(\@address, $netmask), \@network)) {
        exit(0);
    } else {
        exit(1);
    }
}


main();
