#!/bin/sh

set -e

PRINTF=/usr/bin/printf

check_for_program()
{
    if ! command -v "$1" >/dev/null 2>&1; then
        echo "Error: $1 is not in your PATH"
        exit 1
    fi
}

usage()
{
    local script
    script=`basename $0`
    local padsss
    padsss=`$PRINTF '%s\n' "$script" | tr '_a-zA-Z0-9.-' '[ *]'`

    echo "USAGE: $script {[-l|--extents ExtentCount] | [-L|--size Size]}" >&2
    echo "       $padsss -n|--name SnapshotLVName -N|--vgname SnapshotVGName" >&2
    echo "       $padsss [-s|--snapshot] OriginPath" >&2
}


# Pre-flight checks

check_for_program blkid
check_for_program lvcreate
check_for_program lvrename
check_for_program vgimportclone


# Argument processing

dashdash=
snapshot_lvname=
snapshot_vgname=
origin_lvpath=
passthrough_args=

while [ $# -gt 0 ]; do
    if [ x"$1" = x"--" ]; then
        dashdash=1
        shift
        continue
    elif [ -n "$dashdash" -o x"$1" = x"${1#-}" ]; then
        if [ -n "$origin_lvpath" ]; then
            echo "Too many arguments" >&2
            usage
            exit 1
        fi
        origin_lvpath="$1"
        shift
        continue
    fi

    case $1 in
    -l|--extents|-L|--size)
        passthrough_args="$passthrough_args `$PRINTF '%q %q' "$1" "$2"`"
        ;;

    -n|--name)
        snapshot_lvname="$2"
        ;;

    -N|--vgname)
        snapshot_vgname="$2"
        ;;

    -s|--snapshot)
        if [ -n "$origin_lvpath" ]; then
            echo "Too many arguments" >&2
            usage
            exit 1
        fi
        origin_lvpath="$2"
        ;;

    *)
        echo "Unknown option '$1'" >&2
        usage
        exit 1
    esac

    if [ $# -lt 2 ]; then
        echo "Missing value for option '$1' " >&2
        usage
        exit 1
    fi

    shift; shift
done

# Check for required arguments

if [ -z "$origin_lvpath" ]; then
    echo "An origin LV path must be specified" >&2
    usage
    exit 1
fi

if [ -z "$snapshot_lvname" ]; then
    echo "The snapshot LV name (--name) must be specified" >&2
    usage
    exit 1
fi

if [ -z "$snapshot_vgname" ]; then
    echo "The snapshot VG name (--vgname) must be specified" >&2
    usage
    exit 1
fi

# Validate arguments

origin_lvpath="${origin_lvpath#/dev/}"
origin_lvname="${origin_lvpath#*/}"

if $PRINTF '%s\n' "$origin_lvpath" | egrep -q '^opc(-[^/]*)?-vg\.'; then
    echo "The origin LV must not be in an OPC VG" >&2
    exit 2
fi
if [ x"$origin_lvname" = x"${origin_lvname#opc-}" ]; then
    echo "The origin LV name does not indicate it is hosting an OPC VG" >&2
    exit 2
fi

if ! $PRINTF '%s\n' "$snapshot_lvname" | egrep -q '^opc-snap[0-9]*-'; then
    echo "The snapshot LV name (--name) must start with 'opc-snap[NNN]-'" >&2
fi

if ! $PRINTF '%s\n' "$snapshot_vgname" | egrep -q '^opc-snap[0-9]*-vg\.'; then
    echo "The snapshot VG name (--vgname) must start with 'opc-snap[NNN]-vg.'" >&2
fi

snapshot_lvpath="${origin_lvpath%%/*}/$snapshot_lvname"
pending_lvname="opc-pending-snap${snapshot_lvname#opc-snap}"
pending_lvpath="${origin_lvpath%%/*}/$pending_lvname"

if [ ! -e "/dev/$origin_lvpath" ]; then
    echo "The specified origin LV does not exist or is not active"
    exit 2
fi

origin_info=`blkid -o export "/dev/$origin_lvpath"`
eval "$origin_info"

if [ x"$TYPE" != xLVM2_member ]; then
    echo "The specified origin LV does not appear to contain LVM data"
    exit 2
fi

if [ -e "/dev/$snapshot_lvpath" -o -e "/dev/$pending_lvpath" ]; then
    echo "The specified snapshot LV already exists or is pending" >&2
    exit 2
fi

# Perform the steps required to snapshot an OPC user's LVM pool

lvcreate $passthrough_args                      \
         --name "$pending_lvname"               \
         --snapshot "$origin_lvpath"

# For the vgimportclone command only, we need to allow access to pending
# snapshot LVs.  The global_filter in the /etc/lvm/lvm.conf installed on OPC
# hosts normally prevents access.

import_filter='[ "r|/dm-|",
	         "r|/lvm-pv-uuid-|",
	         "r|/opc(--?snap[0-9]*)?--?vg\.|",
	         "a|.*|" ]'

vgimportclone --config "devices { global_filter = $import_filter }"     \
              --basevgname "$snapshot_vgname"                           \
              "/dev/$pending_lvpath" ||
    (rc=$?; lvremove --force "$pending_lvpath"; exit $rc)

lvrename "$pending_lvpath" "$snapshot_lvname" ||
    (rc=$?; lvremove --force "$pending_lvpath"; exit $rc)

exit 0
